
;; Copyright (c) 2012 United States Government as represented by
;; the National Aeronautics and Space Administration.  No copyright
;; is claimed in the United States under Title 17, U.S.Code. All Other
;; Rights Reserved.
;;
;; The software in this package has been released as open-source software
;; under the NASA Open Source Agreement.  See the accompanying file
;; NASA_Open_Source_Agreement.pdf for a full description of the terms.


;; This Elisp code is used to load and execute the PVS theory extraction
;; functions.  It serves as the interface when running PVS in batch mode.

"Extract information from all PVS libraries listing in manifest file using
Lisp-based extraction functions."

(setf *hypatheon-data-dir*   (getenv "HYPATHEON_TEMP_DIR"))
(setf *hypatheon-script-dir* (getenv "HYPATHEON_SCRIPT_DIR"))
(setf *hypatheon-log-dir*    (getenv "HYPATHEON_LOG_DIR"))

(defun do-extraction-notify ()
  (do-extraction "t"))

(defun do-extraction-silent ()
  (do-extraction "nil"))

(defun do-extraction (notify)
   (pvs-message "Invoked do-extraction.")
   (pvs-message (format "PVS_LIBRARY_PATH:  %s" (getenv "PVS_LIBRARY_PATH")))
   ;; Additional strategies must be loaded here

   (let* ((binary-type (pvs-send-and-wait "*pvs-binary-type*"))
	  (lisp-extract-file
	   (concat *hypatheon-script-dir* "/extract-pvs-data.lisp"))
	  (bin-extract-file
	   (concat *hypatheon-script-dir* "/extract-pvs-data." binary-type)))
     (unless (file-newer-than-file-p bin-extract-file lisp-extract-file)
       (pvs-message "Starting to compile extract-pvs-data.lisp ")
       (with-timeout (30 (kill-emacs 3))
	 (print (pvs-send-and-wait
		 (format "(compile-file %S)" lisp-extract-file))))))

   (with-timeout (10 (kill-emacs 3))
      (print (pvs-send-and-wait
	      (concat "(load \"" *hypatheon-script-dir* "/extract-pvs-data\" )"))))
   (pvs-message "Loaded extraction utility.")

   (with-timeout (86400 (kill-emacs 3))    ;; 24 hours
      (let ((status
	     (condition-case ()
		 (pvs-send-and-wait
		  (concat "(extract-all-libraries " notify
			  " \"" *hypatheon-data-dir* "\" " 
			  " \"" *hypatheon-log-dir* "\" " 
			  " \"manifest\")"))
	       ;; Parsing and typecheck errors can generate Emacs-level errors
	       ;; that propagate here.
	       (error 2))))
	(if (equal status 0)
	    (pvs-message "Library extraction completed normally.")
	    (progn (pvs-message "Library extraction terminated prematurely.")
		   (kill-emacs status))))))
